/* -*- pftp-c -*- */

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif
#if HAVE_STDIO_H
# include <stdio.h>
#endif
#if HAVE_INTTYPES_H
# include <inttypes.h>
#endif
#if HAVE_STDLIB_H
# include <stdlib.h>
#endif
#if HAVE_STRING_H
# include <string.h>
#endif

#if HAVE_FAM_H
# include <fam.h>
#endif

#include <gtk/gtk.h>

#include "local_monitor.h"

#ifdef DEBUG
# include <assert.h>
#else
# define assert(x) // x
#endif

#ifdef WITH_DMALLOC
# include <dmalloc.h>
#endif

typedef struct monitor_s *monitor_t;

struct monitor_s {
    local_monitor_callback_t callback;
    void *param;
    local_monitor_id_t id;
    monitor_t next;
    FAMRequest fam;
};

struct local_monitor_s {
    FAMConnection *fam;
    local_monitor_id_t maxid;
    GIOChannel *io;
    guint eid;
    monitor_t first, last;
};

static gboolean _incoming_fam(GIOChannel *source, GIOCondition condition,
			      gpointer data);

local_monitor_t initLocalMonitor(const char *prgname)
{
    local_monitor_t ret;

    ret = malloc(sizeof(struct local_monitor_s));
    memset(ret, 0, sizeof(struct local_monitor_s));

    ret->maxid = 0;
    ret->fam = malloc(sizeof(FAMConnection));
    
    if (FAMOpen2(ret->fam, prgname) != 0) {
	fprintf(stderr, "FAM: Unable to connect to famd.\n");
	free(ret->fam);
	free(ret);
	return NULL;
    }

    if (!(ret->io = g_io_channel_unix_new(FAMCONNECTION_GETFD(ret->fam)))) {
	fprintf(stderr, "gIO: Unable to setup an IO Channel.\n");
	FAMClose(ret->fam);
	free(ret->fam);
	free(ret);
	return NULL;
    }
    if (g_io_channel_set_encoding(ret->io, NULL, NULL) != G_IO_STATUS_NORMAL) {
	fprintf(stderr, "gIO: Unable to set binary encoding.\n");
	g_io_channel_unref(ret->io);
	FAMClose(ret->fam);
	free(ret->fam);
	free(ret);
	return NULL;
    }

    ret->eid = g_io_add_watch(ret->io, G_IO_IN, _incoming_fam, ret);

    return ret;
}

void freeLocalMonitor(local_monitor_t lm)
{
    if (lm) {       
	monitor_t cur;
	if (lm->io)
	    g_io_channel_unref(lm->io);
	if (lm->fam) {
	    FAMClose(lm->fam);
	    free(lm->fam);
	}
	cur = lm->first;
	while (cur) {
	    monitor_t next;
	    next = cur->next;
	    free(cur);
	    cur = next;
	}
	free(lm);
    }
}

local_monitor_id_t monitorDirectory(local_monitor_t lm, const char *fullpath,
				    local_monitor_callback_t callback,
				    void *param)
{
    monitor_t cur;
    cur = malloc(sizeof(struct monitor_s));
    memset(cur, 0, sizeof(struct monitor_s));

    assert(fullpath && callback);

    if (FAMMonitorDirectory(lm->fam, fullpath, &cur->fam, cur)) {
	free(cur);
	return INVALID_LM_ID;
    }

    cur->callback = callback;
    cur->param = param;
    cur->id = lm->maxid++;

    if (lm->last)
	lm->last->next = cur;
    lm->last = cur;

    if (!lm->first)
	lm->first = cur;

    return cur->id;
}

void stopMonitoring(local_monitor_t lm, local_monitor_id_t id)
{
    monitor_t cur;
    cur = lm->first;

    if (id == (lm->maxid - 1))
	lm->maxid--;

    if (!cur) {
	assert(0);
	return;
    }

    if (cur->id == id) {
	lm->first = cur->next;
	FAMCancelMonitor(lm->fam, &cur->fam);
	free(cur);
	return;
    }

    while (cur->next) {
	if (cur->next->id == id) {
	    monitor_t old;
	    old = cur->next;
	    cur->next = cur->next->next;
	    FAMCancelMonitor(lm->fam, &old->fam);
	    free(old);
	    if (cur->next == NULL)
		lm->last = cur;
	    return;
	} else {
	    cur = cur->next;
	}
    }

    assert(0);
}

gboolean _incoming_fam(GIOChannel *source, GIOCondition condition,
		       gpointer data)
{
    local_monitor_t lm;

    lm = (local_monitor_t)data;
    
    while (FAMPending(lm->fam)) {
	FAMEvent ev;
	unsigned int code;
	char *arg;
	code = 0;
	arg = NULL;

	if (FAMNextEvent(lm->fam, &ev) != 1) {
	    fprintf(stderr, "FAM: Failed to read next FAM event. Won\'t try again.\n");
	    return FALSE;
	}
       
	switch (ev.code) {
	case FAMCreated:
	    /* Called when a file is created in the directory we monitor */
	    if (!code) code = LM_CREATED;
	case FAMChanged:
	    /* Called when fstat whould give different data for an object. */
	    if (!code) code = LM_CHANGED;
	case FAMDeleted:
	    /* Called when a file is removed from directory.
	       Can be if directory itself is removed. */
	    if (!code) code = LM_DELETED;

	    if (ev.filename[0] == '/') {
		/* Directory is target */
		arg = strdup(".");
	    } else {
		/* File in directory is target */
		arg = strdup(ev.filename);
	    }
	    break;
	case FAMStartExecuting:
	    /* Ignored. */
	    break;
	case FAMStopExecuting:
	    /* Ignored. */
	    break;
        case FAMMoved:
	    /* Isn't sent anymore. */
	    break;
	case FAMAcknowledge:
	    /* Sent when monitor is canceled or was given invalid filename. */
	    break;
	case FAMExists:
	    /* Sent for the directory and all it's files. */
	    break;
	case FAMEndExist:
	    /* Send when there are no more FAMExists to send. */
	    break;
	}

	if (code && arg) {
	    monitor_t cur;
	    cur = (monitor_t) ev.userdata;

	    if (memcmp(&cur->fam, &ev.fr, sizeof(FAMRequest)) == 0) {
		cur->callback(cur->id, code, arg, cur->param);
		free(arg);
	    }
	}
    }

    return TRUE;
}
